"""Various utilites for analytics
"""
from typing import List, Dict, Any, Tuple
from ..read import ViReadClient


class ViAnalyticsUtils(ViReadClient):
    def sort_documents_by_value(self, documents: List[Dict[str, Any]], sort_by: str, reverse=False):
        """
        Sort documents based on a specific value

        Args:
            documents:
                list of documents.
            sort_by:
                the field by which to sort
        
        Example:
            >>> from vectorai.analytics import ViAnalyticsUtils
            >>> ViAnalyticsUtils.sort_documents_by_values(documents, sort_by='random_field.here')
        """

        docs = sorted(
            documents,
            key=lambda i: self.get_field(sort_by, i),
            reverse=reverse,
        )
        return docs


class MeanDict:
    """
    A special kind of dictionary that lets you take the average of floats as you add them to a dictionary.
    
    Example:
        >>> from vectorai.analytics.utils import MeanDict
        >>> mean_dict = MeanDict()
        >>> mean_dict['a'] = 12
        >>> mean_dict['a'] = 24
        >>> print(mean_dict['a'])
        18
    """

    def __init__(self):
        self._dict = {}
        self._counter = {}

    def __setitem__(self, k: str, item: float):
        """
        What number to add to an item in the meandict.

        Args:
            k:
                Key for mean dict
            item:
                the item value for the key.
        
        Example:
            >>> from vectorai.analytics.utils import MeanDict
            >>> mean_dict = MeanDict()
            >>> mean_dict['a'] = 12
            >>> mean_dict['a'] = 24
            >>> print(mean_dict['a'])
        """
        if k not in self._dict.keys():
            self._dict[k] = item
            self._counter[k] = 1
        else:
            total = self._dict[k] * self._counter[k]
            self._counter[k] = self._counter[k] + 1
            self._dict[k] = (total + item) / self._counter[k]

    def __getitem__(self, k):
        """
        Return the value from the meandict
        Args:
            k:
                Key for mean dict

        Example:
            >>> from vectorai.analytics.utils import MeanDict
            >>> mean_dict = MeanDict()
            >>> mean_dict['a'] = 12
            >>> mean_dict['a'] = 24
            >>> print(mean_dict['a'])
        """
        return self._dict[k]

    def get_x_y(self) -> Tuple[List, List]:
        """
        Return the keys and values as 2 lists.

        Example:
            >>> from vectorai.analytics.utils import MeanDict
            >>> mean_dict = MeanDict()
            >>> mean_dict['a'] = 12
            >>> mean_dict['a'] = 24
            >>> mean_dict.get_x_y()
            ['a'], [18]
        """
        return_dict = {
            k: v for k, v in sorted(self._dict.items(), key=lambda item: item[1])
        }
        return list(return_dict.keys()), list(return_dict.values())



"""Base class for models
"""
from abc import ABC, abstractmethod
from typing import List, Dict, Union
import json
import gc
import os
from appdirs import *
from ..client import ViWriteClient

class _Vi2Vec(ViWriteClient):
    """
    Abstract class for text models. We inherit from the base class to make use of the ability to write 
    to nested dictionaries and other utilities that will be helpful.
    """

    def chunk(self, documents, chunk_size=20):
        return ViWriteClient.chunk(documents, chunk_size)

    def init_tpu(self):
        """
            Initialise TPU for training a model.

            Example::

                >>> from vectorai.client import Transformer2Vec
                >>> model = Transformer2Vec
                >>> model.init_tpu()
        """
        try:
            tpu = tf.distribute.cluster_resolver.TPUClusterResolver()  # TPU detection
            print("Running on TPU ", tpu.cluster_spec().as_dict()["worker"])
            IS_TPU = True
            tf.config.experimental_connect_to_cluster(tpu)
            tf.tpu.experimental.initialize_tpu_system(tpu)
            self.tpu_strategy = tf.distribute.experimental.TPUStrategy(tpu)
        except:
            pass
    
    @staticmethod
    def is_json_serializable(document: dict):
        """
            Check to ensure a document is JSON serializable.
            
            Example::

                >>> from vectorai.client import Transformer2Vec
                >>> sample_doc = {'test': 'hi'}
                >>> Transformer2Vec.is_json_serializable(sample_doc)
                >>> model.init_tpu()
        """
        assert isinstance(document, dict), "Check only 1 item!"
        test_json = user_data_dir("test_vectorai", "test.json")

        with open(test_json, "w"):
            json.dump(test_json)

        with open(test_json) as f:
            check_json = json.load(f)
        os.remove(test_json)
        
        assert check_json == document, (
            "This will not upload correctly. Please ensure all items "
            + "in the dictionary are lists/floats/ints/strings."
        )
        print("Checked! Feel free to upload!")

    @abstractmethod
    def encode(x):
        """An abstract method to specify the encode method.
        """
        pass


class ViText2Vec(_Vi2Vec):
    """Abstract class for text models.
    """

    @abstractmethod
    def encode_text(self, text: Union[str, Dict]) -> List[float]:
        """Encodes either string or a document.
        """
        pass

    def encode(self, text: str):
        """
            Text to vector.
            
            Args:
                text:
                    Text to encode.
            
            Example::

                >>> from vectorai.client import Transformer2Vec
                >>> sample_doc = {'test': 'hi'}
                >>> model_transformer = Transformer2Vec('distilbert')
                >>> model_transformer.encode("Riemann Sums.")
        """
        return self.encode_text(text=text)

    def bulk_encode(self, *args, **kwargs):
        return self.bulk_encode_text(*args, **kwargs)

    def bulk_encode_text(
        self,
        documents: Union[List[str], List[Dict]],
        document_fields: str = None,
        vector_output_field: str = None,
        chunk_size: int = 10,
    ) -> List[float]:
        """
            Encodes either a list of strings or a list of documents. This can be over-ridden if bulk-encoding is supported
            outside of list comprehension.
            Currently only supports 1 input text field.
            Bulk encoding text assumes that bulk-encoding is best done via list comprehension of encode_text.

            Args:
                documents:
                    A list of Python dictionaries
                document_fields
                    The fields to encode
                vector_output_field:
                    The name of the vector output
                chunk_size (batch size): 
                    The number of documents to encode.

            Example::

                >>> from vectorai.client import Transformer2Vec
                >>> sample_docs = [{'name': 'bert', 'age": 10}, {'name': 'elmo', 'age': 15}]
                >>> model_transformer = Transformer2Vec('distilbert')
                >>> model_transformer.bulk_encode(sample_docs, "name")
        
        """
        if vector_output_field is None:
            vector_output_field = document_fields + '_vector_'
        
        if isinstance(documents[0], str):
            assert (
                document_fields is None
            ), "You cannot have an input text field if you are just feeding in a list of documents."
            assert (
                vector_output_field is None
            ), "You cannot have a vector output field if you are feeding i- Overfitting does not seem to provide good results. "
            # typechecking to ensure correct input is fed in.
            all_vectors = []
            for chunk in self.chunk(documents, chunk_size=chunk_size):
                all_vectors.append(self.encode_text(chunk))
            return all_vectors

        if isinstance(documents[0], dict):
            assert (
                document_fields is not None
            ), "You need a text input field if you are not feeding in at a document level."
            assert (
                vector_output_field is not None
            ), "You need a vector output field if you are feeding in at a document level."
            for chunk in self.chunk(documents, chunk_size=chunk_size):
                # List comprehension of the text encoding methodology.
                vectors = [
                    self.encode_text(document=x, document_fields=document_fields)
                    for x in chunk
                ]
                [
                    x.update({vector_output_field: vector})
                    for x, vector in list(zip(chunk, vectors))
                ]
                del vectors
                gc.collect()
            print("Finished updating documents with additional field.")
            return
        raise ValueError(
            "Unsure of how to bulk encode. Please write custom encoding methodology"
        )


class ViImage2Vec(_Vi2Vec):
    """Abstract class for image models.
    """

    @abstractmethod
    def encode_image(self, image) -> List[float]:
        pass

    @abstractmethod
    def bulk_encode_image(self, images):
        pass


class ViAudio2Vec(_Vi2Vec):
    """Abstract class for audio models.
    """

    @abstractmethod
    def encode_audio(self, audio) -> List[float]:
        pass

    @abstractmethod
    def bulk_encode_audio(self, audio):
        pass



import os
from .comparator import ComparatorClient
from .viz import VizMixin

class ViAnalyticsClient(ComparatorClient, VizMixin):
    def __init__(self, username: str=None, api_key: str=None,
    url: str = "https://api.vctr.ai", analytics_url="https://vector-analytics.vctr.ai"):
        self.username = username if username is not None else os.environ['VI_USERNAME']
        self.api_key = api_key if api_key is not None else os.environ['VI_API_KEY']
        self.url = url
        self.analytics_url = analytics_url



import requests
from typing import List, Dict, Optional
from ...api.utils import retry, return_curl_or_response

class ComparatorAPI:
    def __init__(self, username: str=None, api_key: str=None, 
    url: str = "https://api.vctr.ai", analytics_url="https://vector-analytics.vctr.ai"):
        self.username = username
        self.api_key = api_key
        self.url = url
        self.analytics_url = analytics_url


    @return_curl_or_response('content')
    @retry()
    def _compare_ranks(
        self, 
        ranked_list_1: List[Dict],
        ranked_list_2: List[Dict],
        fields_to_display: List[str]=None,
        image_fields: List[str]=[],
        audio_fields: List[str]=[],
        column_titles: Optional[List[str]] = None,
        x_axis_title: str = 'Fields',
        y_axis_title: str = 'Comparing: ',
        header: str = "<h1>Top-K Ranking Comparator</h1>",
        subheader: str = "<h2>Compare ranks in the different lists.</h2>",
        colors: List[str]=['#ccff99', 'powderblue', '#ffc2b3'],
        return_curl: bool=False,
        **kwargs
        ):
        """
        Compare Top-K Lists.
        Args:
            ranked_list_1: A list of results as a dictionary containing the required fields.
            ranked_list_2: Another list of results
            fields_to_display: The fields required for displaying the object
            image_fields: The fields which are images 
            audio_fields: The fields which are audio
            column_titles: The name of the columns for the differnt rank fields
            x_axis_title: The title of the x axis 
            y_axis_title: The title of the y axis
            header: The name of the graph 
            subheader: The sub-header of the graph
        """

        params={
            "username": self.username,
            "api_key": self.api_key,
            "ranked_list_1": ranked_list_1,
            "ranked_list_2": ranked_list_2,
            "fields_to_display": fields_to_display,
            "image_fields": image_fields,
            "audio_fields": audio_fields,
            "column_titles": column_titles,
            "x_axis_title": x_axis_title,
            "y_axis_title": y_axis_title,
            "header": header,
            "subheader": subheader,
            "colors": colors,
        }
        params.update(kwargs)
        return requests.post(
            url= f"{self.analytics_url}/comparator/compare_ranks/",
            json=params)



# This python file is auto-generated. Please do not edit.
import requests
import requests
from .utils import retry, return_curl_or_response


class _ViAPIClient:
	def __init__(self, username, api_key, url, ):
		self.username = username		
		self.api_key = api_key		
		self.url = url		

	@retry()
	@return_curl_or_response('json')
	def request_api_key(self, email, description, referral_code="api_referred", **kwargs):
		"""Request an api key
Make sure to save the api key somewhere safe. If you have a valid referral code, you can recieve the api key more quickly.
    
Args
========
username: Username you'd like to create, lowercase only
email: Email you are using to sign up
description: Description of your intended use case
referral_code: The referral code you've been given to allow you to register for an api key before others

"""
		return requests.post(
			url=self.url+'/project/request_api_key',
			json=dict(
				username=self.username,
				email=email, 
				description=description, 
				referral_code=referral_code, 
				))

	@retry()
	@return_curl_or_response('json')
	def request_read_api_key(self, read_username, **kwargs):
		"""Request a read api key for your collections
Creates a read only key for your collections. Make sure to save the api key somewhere safe. When doing a search the admin username should still be used.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
read_username: Username for read only key

"""
		return requests.post(
			url=self.url+'/project/request_read_api_key',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				read_username=read_username, 
				))

	@retry()
	@return_curl_or_response('json')
	def _create_collection(self, collection_name, collection_schema={}, **kwargs):
		"""Creates a collection
A collection can store documents to be **searched, retrieved, filtered and aggregated** _(similar to Collections in MongoDB, Tables in SQL, Indexes in ElasticSearch)_.

If you are inserting your own vector use the suffix (ends with) **"\_vector\_"** for the field name. and specify the length of the vector in colletion_schema like below example:

    {
        "collection_schema": {
            "celebrity_image_vector_": 1024,
            "celebrity_audio_vector" : 512,
            "product_description_vector" : 128
        }
    }
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
collection_schema: Schema for specifying the field that are vectors and its length

"""
		return requests.post(
			url=self.url+'/project/create_collection',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				collection_schema=collection_schema, 
				))

	@retry()
	@return_curl_or_response('json')
	def _create_collection_from_document(self, collection_name, document={}, **kwargs):
		"""Creates a collection by infering the schema from a document
If you are inserting your own vector use the suffix (ends with)  **"\_vector\_"** for the field name. e.g. "product\_description\_vector\_"
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
document: A Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'

"""
		return requests.post(
			url=self.url+'/project/create_collection_from_document',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				document=document, 
				))

	@retry()
	@return_curl_or_response('json')
	def _delete_collection(self,collection_name, **kwargs):
		return requests.get(
			url=self.url+'/project/delete_collection',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def _list_collections(self,**kwargs):
		return requests.get(
			url=self.url+'/project/list_collections',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				))

	@retry()
	@return_curl_or_response('json')
	def collection_stats(self,collection_name, seperate_chunks=False, **kwargs):
		return requests.get(
			url=self.url+'/project/collection_stats',
			params=dict(
				seperate_chunks=seperate_chunks, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def collection_schema(self,collection_name, **kwargs):
		return requests.get(
			url=self.url+'/project/collection_schema',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def collection_vector_health(self,collection_name, **kwargs):
		return requests.get(
			url=self.url+'/project/collection_vector_health',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def add_collection_metadata(self, collection_name, metadata, **kwargs):
		"""Add metadata about a collection
Add metadata about a collection. notably description, data source, etc
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
metadata: Metadata for a collection, e.g. {'description' : 'collection for searching products'}

"""
		return requests.post(
			url=self.url+'/project/add_collection_metadata',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				metadata=metadata, 
				))

	@retry()
	@return_curl_or_response('json')
	def collection_metadata(self,collection_name, **kwargs):
		return requests.get(
			url=self.url+'/project/collection_metadata',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def copy_collection(self, collection_name, original_collection_name, collection_schema={}, rename_fields={}, remove_fields=[], **kwargs):
		"""Copy a collection into a new collection
Copy a collection into a new collection. You can use this to rename fields and change data schema.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
original_collection_name: Name of collection to copy from
collection_schema: Schema to change, if unspecified then schema is unchanged. Defaults to no schema change
rename_fields: Fields to rename {'old_field': 'new_field'}. Defaults to no renames
remove_fields: Fields to remove ['random_field', 'another_random_field']. Defaults to no removes

"""
		return requests.post(
			url=self.url+'/project/copy_collection',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				original_collection_name=original_collection_name, 
				collection_schema=collection_schema, 
				rename_fields=rename_fields, 
				remove_fields=remove_fields, 
				))

	@retry()
	@return_curl_or_response('json')
	def job_status(self,job_id, **kwargs):
		return requests.get(
			url=self.url+'/project/job_status',
			params=dict(
				job_id=job_id, 
				username=self.username, 
				api_key=self.api_key, 
				))

	@retry()
	@return_curl_or_response('json')
	def insert(self, collection_name, document={}, insert_date=True, overwrite=True, update_schema=True, quick=False, **kwargs):
		"""Insert a document into a Collection
When inserting the document you can specify your own id for a document by using the field name **"\_id"**. 
For specifying your own vector use the suffix (ends with)  **"\_vector\_"** for the field name.
e.g. "product\_description\_vector\_"
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
document: A Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'
insert_date: Whether to include insert date as a field 'insert_date_'.
overwrite: Whether to overwrite document if it exists.
update_schema: Whether the api should check the documents for vector datatype to update the schema.
quick: This will run the quickest insertion possible, which means there will be no schema checks or collection checks.

"""
		return requests.post(
			url=self.url+'/collection/insert',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				document=document, 
				insert_date=insert_date, 
				overwrite=overwrite, 
				update_schema=update_schema, 
				quick=quick, 
				))

	@retry()
	@return_curl_or_response('json')
	def bulk_insert(self, collection_name, documents={}, insert_date=True, overwrite=True, update_schema=True, quick=False, **kwargs):
		"""Insert multiple documents into a Collection
When inserting the document you can specify your own id for a document by using the field name **"\_id"**. 
For specifying your own vector use the suffix (ends with)  **"\_vector\_"** for the field name.
e.g. "product\_description\_vector\_"
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
documents: A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'
insert_date: Whether to include insert date as a field 'insert_date_'.
overwrite: Whether to overwrite document if it exists.
update_schema: Whether the api should check the documents for vector datatype to update the schema.
quick: This will run the quickest insertion possible, which means there will be no schema checks or collection checks.

"""
		return requests.post(
			url=self.url+'/collection/bulk_insert',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				documents=documents, 
				insert_date=insert_date, 
				overwrite=overwrite, 
				update_schema=update_schema, 
				quick=quick, 
				))

	@retry()
	@return_curl_or_response('json')
	def delete_by_id(self,document_id, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/delete_by_id',
			params=dict(
				document_id=document_id, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def bulk_delete_by_id(self, collection_name, document_ids, **kwargs):
		"""Delete multiple documents in a Collection by ids
Delete a document by its id.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
document_ids: IDs of documents

"""
		return requests.post(
			url=self.url+'/collection/bulk_delete_by_id',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				document_ids=document_ids, 
				))

	@retry()
	@return_curl_or_response('json')
	def edit_document(self, collection_name, document_id, edits, insert_date=True, **kwargs):
		"""Edit a document in a Collection by its id
Edit by providing a key value pair of fields you are adding or changing.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
document_id: ID of a document
edits: A dictionary to edit and add fields to a document.
insert_date: Whether to include insert date as a field 'insert_date_'.

"""
		return requests.post(
			url=self.url+'/collection/edit_document',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				document_id=document_id, 
				edits=edits, 
				insert_date=insert_date, 
				))

	@retry()
	@return_curl_or_response('json')
	def bulk_edit_document(self, collection_name, documents={}, insert_date=True, **kwargs):
		"""Edits multiple documents in a Collection by its ids
Edits documents by providing a key value pair of fields you are adding or changing, make sure to include the "_id" in the documents.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
documents: A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '\_id', for specifying vector field use the suffix of '\_vector\_'
insert_date: Whether to include insert date as a field 'insert_date_'.

"""
		return requests.post(
			url=self.url+'/collection/bulk_edit_document',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				documents=documents, 
				insert_date=insert_date, 
				))

	@retry()
	@return_curl_or_response('json')
	def delete_document_fields(self,document_id, fields_to_delete, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/delete_document_fields',
			params=dict(
				document_id=document_id, 
				fields_to_delete=fields_to_delete, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def id(self,document_id, collection_name, include_vector=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/id',
			params=dict(
				document_id=document_id, 
				include_vector=include_vector, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def bulk_id(self,document_ids, collection_name, include_vector=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/bulk_id',
			params=dict(
				document_ids=document_ids, 
				include_vector=include_vector, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def bulk_missing_id(self, collection_name, document_ids, **kwargs):
		"""Look up in bulk if the ids exists in the collection, returns all the missing one as a list
Look up in bulk if the ids exists in the collection, returns all the missing one as a list.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
document_ids: IDs of documents

"""
		return requests.post(
			url=self.url+'/collection/bulk_missing_id',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				document_ids=document_ids, 
				))

	@retry()
	@return_curl_or_response('json')
	def retrieve_documents(self,collection_name, include_fields=[], cursor=None, page_size=20, sort=[], asc=False, include_vector=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/retrieve_documents',
			params=dict(
				include_fields=include_fields, 
				cursor=cursor, 
				page_size=page_size, 
				sort=sort, 
				asc=asc, 
				include_vector=include_vector, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def random_documents(self,collection_name, seed=10, include_fields=[], page_size=20, include_vector=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/random_documents',
			params=dict(
				seed=seed, 
				include_fields=include_fields, 
				page_size=page_size, 
				include_vector=include_vector, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def retrieve_documents_with_filters(self, collection_name, include_fields=[], cursor=None, page_size=20, sort=[], asc=False, include_vector=True, filters=[], **kwargs):
		"""Retrieve some documents with filters
Cursor is provided to retrieve even more documents. Loop through it to retrieve all documents in the database.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
include_fields: Fields to include in the search results, empty array/list means all fields.
cursor: Cursor to paginate the document retrieval
page_size: Size of each page of results
sort: Fields to sort by
asc: Whether to sort results by ascending or descending order
include_vector: Include vectors in the search results
filters: Query for filtering the search results

"""
		return requests.post(
			url=self.url+'/collection/retrieve_documents_with_filters',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				include_fields=include_fields, 
				cursor=cursor, 
				page_size=page_size, 
				sort=sort, 
				asc=asc, 
				include_vector=include_vector, 
				filters=filters, 
				))

	@retry()
	@return_curl_or_response('json')
	def random_documents_with_filters(self, collection_name, seed=10, include_fields=[], page_size=20, include_vector=True, filters=[], **kwargs):
		"""Retrieve some documents randomly with filters
Mainly for testing purposes.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
seed: Random Seed for retrieving random documents.
include_fields: Fields to include in the search results, empty array/list means all fields.
page_size: Size of each page of results
include_vector: Include vectors in the search results
filters: Query for filtering the search results

"""
		return requests.post(
			url=self.url+'/collection/random_documents_with_filters',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				seed=seed, 
				include_fields=include_fields, 
				page_size=page_size, 
				include_vector=include_vector, 
				filters=filters, 
				))

	@retry()
	@return_curl_or_response('json')
	def _search(self,vector, collection_name, search_fields, approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/search',
			params=dict(
				vector=vector, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_fields=search_fields, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_by_id(self,document_id, collection_name, search_field, approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/search_by_id',
			params=dict(
				document_id=document_id, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_field=search_field, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_by_ids(self,document_ids, collection_name, search_field, vector_operation="sum", approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/search_by_ids',
			params=dict(
				document_ids=document_ids, 
				vector_operation=vector_operation, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_field=search_field, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_by_positive_negative_ids(self,positive_document_ids, negative_document_ids, collection_name, search_field, vector_operation="sum", approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/search_by_positive_negative_ids',
			params=dict(
				positive_document_ids=positive_document_ids, 
				negative_document_ids=negative_document_ids, 
				vector_operation=vector_operation, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_field=search_field, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_with_positive_negative_ids_as_history(self,vector, positive_document_ids, negative_document_ids, collection_name, search_field, vector_operation="sum", approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/search_with_positive_negative_ids_as_history',
			params=dict(
				vector=vector, 
				positive_document_ids=positive_document_ids, 
				negative_document_ids=negative_document_ids, 
				vector_operation=vector_operation, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_field=search_field, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def facets(self,collection_name, facets_fields=[], date_interval="monthly", page_size=1000, page=1, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/facets',
			params=dict(
				facets_fields=facets_fields, 
				date_interval=date_interval, 
				page_size=page_size, 
				page=page, 
				asc=asc, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def filters(self, collection_name, filters=[], page=1, page_size=20, asc=False, include_vector=False, sort=[], **kwargs):
		"""Filters a collection
Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched.

The filters query is a json body that follows the schema of:

    [
        {'field' : <field to filter>, 'filter_type' : <type of filter>, "condition":"==", "condition_value":"america"},
        {'field' : <field to filter>, 'filter_type' : <type of filter>, "condition":">=", "condition_value":90},
    ]

These are the available filter_type types:

1. "contains": for filtering documents that contains a string.
        {'field' : 'category', 'filter_type' : 'contains', "condition":"==", "condition_value": "bluetoo"]}
2. "exact_match"/"category": for filtering documents that matches a string or list of strings exactly.
        {'field' : 'category', 'filter_type' : 'category', "condition":"==", "condition_value": "tv"]}
3. "categories": for filtering documents that contains any of a category from a list of categories.
        {'field' : 'category', 'filter_type' : 'categories', "condition":"==", "condition_value": ["tv", "smart", "bluetooth_compatible"]}
4. "exists": for filtering documents that contains a field.
        {'field' : 'purchased', 'filter_type' : 'exists', "condition":">=", "condition_value":" "}
5. "date": for filtering date by date range.
        {'field' : 'insert_date_', 'filter_type' : 'date', "condition":">=", "condition_value":"2020-01-01"}
6. "numeric": for filtering by numeric range. 
        {'field' : 'price', 'filter_type' : 'numeric', "condition":">=", "condition_value":90}
7. "ids": for filtering by document ids. 
        {'field' : 'ids', 'filter_type' : 'ids', "condition":"==", "condition_value":["1", "10"]}

These are the available conditions:
 
    "==", "!=", ">=", ">", "<", "<="
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
filters: Query for filtering the search results
page: Page of the results
page_size: Size of each page of results
asc: Whether to sort results by ascending or descending order
include_vector: Include vectors in the search results
sort: Fields to sort by

"""
		return requests.post(
			url=self.url+'/collection/filters',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				filters=filters, 
				page=page, 
				page_size=page_size, 
				asc=asc, 
				include_vector=include_vector, 
				sort=sort, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_search(self, collection_name, multivector_query, page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_fields=[], include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, **kwargs):
		"""Advanced Vector Similarity Search. Support for multiple vectors, vector weightings, facets and filtering
Advanced Vector Similarity Search, enables machine learning search with vector search. Search with a multiple vectors for the most similar documents.

For example: Search with a product image and description vectors to find the most similar products by what it looks like and what its described to do.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
multivector_query: Query for advance search that allows for multiple vector and field querying

"""
		return requests.post(
			url=self.url+'/collection/advanced_search',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				multivector_query=multivector_query, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_search_by_id(self, collection_name, document_id, search_fields, page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_fields=[], include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, **kwargs):
		"""Advanced Single Product Recommendations
Single Product Recommendations (Search by an id).

For example: Search with id of a product in the database, and using the product's image and description vectors to find the most similar products by what it looks like and what its described to do.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
document_id: ID of a document
search_fields: Vector fields to search against, and the weightings for them.

"""
		return requests.post(
			url=self.url+'/collection/advanced_search_by_id',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				document_id=document_id, 
				search_fields=search_fields, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_search_by_ids(self, collection_name, document_ids, search_fields, page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_fields=[], include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, vector_operation="sum", **kwargs):
		"""Advanced Multi Product Recommendations
Advanced Multi Product Recommendations (Search by ids).

For example: Search with multiple ids of products in the database, and using the product's image and description vectors to find the most similar products by what it looks like and what its described to do.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

You can also give weightings of on each product as well e.g. product ID-A weights 100% whilst product ID-B 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
document_ids: Document IDs to get recommendations for, and the weightings of each document
search_fields: Vector fields to search against, and the weightings for them.
vector_operation: Aggregation for the vectors, choose from ['mean', 'sum', 'min', 'max']

"""
		return requests.post(
			url=self.url+'/collection/advanced_search_by_ids',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				document_ids=document_ids, 
				search_fields=search_fields, 
				vector_operation=vector_operation, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_search_by_positive_negative_ids(self, collection_name, positive_document_ids, negative_document_ids, search_fields, page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_fields=[], include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, vector_operation="sum", **kwargs):
		"""Advanced Multi Product Recommendations with likes and dislikes
Advanced Multi Product Recommendations with Likes and Dislikes (Search by ids).

For example: Search with multiple ids of liked and dislike products in the database. Then using the product's image and description vectors to find the most similar products by what it looks like and what its described to do against the positives and most disimilar products for the negatives.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

You can also give weightings of on each product as well e.g. product ID-A weights 100% whilst product ID-B 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
positive_document_ids: Positive Document IDs to get recommendations for, and the weightings of each document
negative_document_ids: Negative Document IDs to get recommendations for, and the weightings of each document
search_fields: Vector fields to search against, and the weightings for them.
vector_operation: Aggregation for the vectors, choose from ['mean', 'sum', 'min', 'max']

"""
		return requests.post(
			url=self.url+'/collection/advanced_search_by_positive_negative_ids',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				positive_document_ids=positive_document_ids, 
				negative_document_ids=negative_document_ids, 
				search_fields=search_fields, 
				vector_operation=vector_operation, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_search_with_positive_negative_ids_as_history(self, collection_name, multivector_query, positive_document_ids, negative_document_ids, page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_fields=[], include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, vector_operation="sum", **kwargs):
		"""Advanced Search with Likes and Dislikes as history
For example: Vector search of a query vector with multiple ids of liked and dislike products in the database. Then using the product's image and description vectors to find the most similar products by what it looks like and what its described to do against the positives and most disimilar products for the negatives.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

You can also give weightings of on each product as well e.g. product ID-A weights 100% whilst product ID-B 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
multivector_query: Query for advance search that allows for multiple vector and field querying
positive_document_ids: Positive Document IDs to get recommendations for, and the weightings of each document
negative_document_ids: Negative Document IDs to get recommendations for, and the weightings of each document
vector_operation: Aggregation for the vectors, choose from ['mean', 'sum', 'min', 'max']

"""
		return requests.post(
			url=self.url+'/collection/advanced_search_with_positive_negative_ids_as_history',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				multivector_query=multivector_query, 
				positive_document_ids=positive_document_ids, 
				negative_document_ids=negative_document_ids, 
				vector_operation=vector_operation, 
				))

	@retry()
	@return_curl_or_response('json')
	def aggregate(self, collection_name, aggregation_query, page_size=20, page=1, asc=False, flatten=True, **kwargs):
		"""Aggregate a collection
Aggregation/Groupby of a collection using an aggregation query.
The aggregation query is a json body that follows the schema of:

    {
        "groupby" : [
            {"name": <nickname/alias>, "field": <field in the collection>, "agg": "category"},
            {"name": <another_nickname/alias>, "field": <another groupby field in the collection>, "agg": "category"}
        ], 
        "metrics" : [
            {"name": <nickname/alias>, "field": <numeric field in the collection>, "agg": "avg"}
        ]
    }
- "groupby" is the fields you want to split the data into. These are the available groupby types:
    - category" : groupby a field that is a category
- "metrics" is the fields you want to metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types: 
    - "avg", "max", "min", "sum", "cardinality"
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
aggregation_query: Aggregation query to aggregate data
page_size: Size of each page of results
page: Page of the results
asc: Whether to sort results by ascending or descending order
flatten: 

"""
		return requests.post(
			url=self.url+'/collection/aggregate',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				aggregation_query=aggregation_query, 
				page_size=page_size, 
				page=page, 
				asc=asc, 
				flatten=flatten, 
				))

	@retry()
	@return_curl_or_response('json')
	def traditional_search(self,collection_name, text, text_fields, fuzzy=-1, join=True, page_size=20, page=1, include_fields=[], include_vector=False, include_count=True, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/traditional_search',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				text=text, 
				text_fields=text_fields, 
				fuzzy=fuzzy, 
				join=join, 
				page_size=page_size, 
				page=page, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def hybrid_search(self,text, vector, text_fields, collection_name, search_fields, traditional_weight=0.075, fuzzy=-1, join=True, approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/hybrid_search',
			params=dict(
				text=text, 
				vector=vector, 
				text_fields=text_fields, 
				traditional_weight=traditional_weight, 
				fuzzy=fuzzy, 
				join=join, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_fields=search_fields, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_hybrid_search(self, collection_name, multivector_query, text, page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_fields=[], include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, text_fields=[], traditional_weight=0.075, fuzzy=-1, join=True, **kwargs):
		"""Advanced Search a text field with vector and text using Vector Search and Traditional Search
Advanced Vector similarity search + Traditional Fuzzy Search with text and vector.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
multivector_query: Query for advance search that allows for multiple vector and field querying
text: Text Search Query (not encoded as vector)
text_fields: Text fields to search against
traditional_weight: Multiplier of traditional search. A value of 0.025~0.1 is good.
fuzzy: Fuzziness of the search. A value of 1-3 is good. For automated fuzzines use -1.
join: Whether to consider cases where there is a space in the word. E.g. Go Pro vs GoPro.

"""
		return requests.post(
			url=self.url+'/collection/advanced_hybrid_search',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				multivector_query=multivector_query, 
				text=text, 
				text_fields=text_fields, 
				traditional_weight=traditional_weight, 
				fuzzy=fuzzy, 
				join=join, 
				))

	@retry()
	@return_curl_or_response('json')
	def job_status(self,job_id, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/job_status',
			params=dict(
				job_id=job_id, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def list_jobs(self,collection_name, show_active_only=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/list_jobs',
			params=dict(
				show_active_only=show_active_only, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_image_field(self, collection_name, task, field, image_field, **kwargs):
		"""Start job to encode image field
Encode image field
    
Args
========
username: 
api_key: 
collection_name: 
task: 
field: 
image_field: 

"""
		return requests.post(
			url=self.url+'/collection/encode_image_field',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				task=task, 
				field=field, 
				image_field=image_field, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_field(self, collection_name, task, field, image_field, **kwargs):
		"""Start job to encode image field
Encode image field
    
Args
========
username: 
api_key: 
collection_name: 
task: 
field: 
image_field: 

"""
		return requests.post(
			url=self.url+'/collection/encode_field',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				task=task, 
				field=field, 
				image_field=image_field, 
				))

	@retry()
	@return_curl_or_response('json')
	def copy_collection_from_another_user(self, collection_name, source_collection_name, source_username, source_api_key, **kwargs):
		"""Copy a collection from another user's projects into your project
Copy a collection from another user's projects into your project. This is considered a project job
Args
========
collection_name: Collection to copy into
username: Your username
api_key: Your api key to access the username
source_collection_name: Collection to copy frpm
source_username: Source username of whom the collection belongs to
source_api_key: Api key to access the source username

"""
		return requests.post(
			url=self.url+'/project/copy_collection_from_another_user',
			json=dict(
				collection_name=collection_name, 
				username=self.username,
				api_key=self.api_key,
				source_collection_name=source_collection_name, 
				source_username=source_username, 
				source_api_key=source_api_key, 
				))

	@retry()
	@return_curl_or_response('json')
	def chunk_search(self, collection_name, chunk_field, vector, search_fields, chunk_scoring="max", page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, **kwargs):
		"""Vector Similarity Search on Chunks.
Vector Similarity Search on chunks.

For example: Search with a person's characteristics, who are the most similar (querying the "persons_characteristics_vector" field):

    Query person's characteristics as a vector: 
    [180, 40, 70] representing [height, age, weight]

    Search Results:
    [
        {"name": Adam Levine, "persons_characteristics_vector" : [180, 56, 71]},
        {"name": Brad Pitt, "persons_characteristics_vector" : [180, 56, 65]},
    ...]
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
chunk_field: Field that the array of chunked documents are.
chunk_scoring: Scoring method for determining for ranking between document chunks.
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
vector: Vector, a list/array of floats that represents a piece of data
search_fields: Vector fields to search against

"""
		return requests.post(
			url=self.url+'/collection/chunk_search',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				chunk_field=chunk_field, 
				chunk_scoring=chunk_scoring, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				vector=vector, 
				search_fields=search_fields, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_chunk_search(self, collection_name, chunk_field, multivector_query, chunk_scoring="max", page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, **kwargs):
		"""Advanced Vector Similarity Search on Chunks. Support for multiple vectors, vector weightings, facets and filtering
Advanced Vector Similarity Search, enables machine learning search with vector search. Search with a multiple vectors for the most similar documents.

For example: Search with a product image and description vectors to find the most similar products by what it looks like and what its described to do.

You can also give weightings of each vector field towards the search, e.g. image\_vector\_ weights 100%, whilst description\_vector\_ 50%.

Advanced search also supports filtering to only search through filtered results and facets to get the overview of products available when a minimum score is set.

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
chunk_field: Field that the array of chunked documents are.
chunk_scoring: Scoring method for determining for ranking between document chunks.
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
multivector_query: Query for advance search that allows for multiple vector and field querying

"""
		return requests.post(
			url=self.url+'/collection/advanced_chunk_search',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				chunk_field=chunk_field, 
				chunk_scoring=chunk_scoring, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				multivector_query=multivector_query, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_multistep_chunk_search(self, collection_name, chunk_field, first_step_multivector_query, chunk_step_multivector_query, chunk_scoring="max", page=1, page_size=20, approx=0, sum_fields=True, metric="cosine", filters=[], facets=[], min_score=None, include_vector=False, include_count=True, include_facets=False, hundred_scale=False, asc=False, first_step_page=1, first_step_page_size=20, **kwargs):
		"""Vector Similarity Search on Chunks.
Vector Similarity Search on chunks.

For example: Search with a person's characteristics, who are the most similar (querying the "persons_characteristics_vector" field):

    Query person's characteristics as a vector: 
    [180, 40, 70] representing [height, age, weight]

    Search Results:
    [
        {"name": Adam Levine, "persons_characteristics_vector" : [180, 56, 71]},
        {"name": Brad Pitt, "persons_characteristics_vector" : [180, 56, 65]},
    ...]
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
chunk_field: Field that the array of chunked documents are.
chunk_scoring: Scoring method for determining for ranking between document chunks.
page: Page of the results
page_size: Size of each page of results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
filters: Query for filtering the search results
facets: Fields to include in the facets, if [] then all
min_score: Minimum score for similarity metric
include_vector: Include vectors in the search results
include_count: Include count in the search results
include_facets: Include facets in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
first_step_multivector_query: Query for advance search that allows for multiple vector and field querying
chunk_step_multivector_query: Query for advance search that allows for multiple vector and field querying
first_step_page: Page of the results
first_step_page_size: Size of each page of results

"""
		return requests.post(
			url=self.url+'/collection/advanced_multistep_chunk_search',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				chunk_field=chunk_field, 
				chunk_scoring=chunk_scoring, 
				page=page, 
				page_size=page_size, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				filters=filters, 
				facets=facets, 
				min_score=min_score, 
				include_vector=include_vector, 
				include_count=include_count, 
				include_facets=include_facets, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				first_step_multivector_query=first_step_multivector_query, 
				chunk_step_multivector_query=chunk_step_multivector_query, 
				first_step_page=first_step_page, 
				first_step_page_size=first_step_page_size, 
				))

	@retry()
	@return_curl_or_response('json')
	def id_lookup_joined(self, join_query, doc_id, **kwargs):
		"""Look up a document by its id with joins
Look up a document by its id with joins.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
join_query: 
doc_id: ID of a Document

"""
		return requests.post(
			url=self.url+'/collection/id_lookup_joined',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				join_query=join_query, 
				doc_id=doc_id, 
				))

	@retry()
	@return_curl_or_response('json')
	def join_collections(self, join_query, joined_collection_name, **kwargs):
		"""Join collections with a query
Perform a join query on a whole collection and write the results to a new collection. We currently only support left joins.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
join_query: 
joined_collection_name: Name of the new collection that contains the joined results

"""
		return requests.post(
			url=self.url+'/collection/join_collections',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				join_query=join_query, 
				joined_collection_name=joined_collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def publish_aggregation(self, source_collection, dest_collection, aggregation_name, description, aggregation_query, date_field="insert_date_", refresh_time="160s", start_immediately=True, **kwargs):
		"""Publishes your aggregation query to a new collection
Publish and schedules your aggregation query and saves it to a new collection.
This new collection is just like any other collection and you can read, filter and aggregate it.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
source_collection: The collection where the data to aggregate comes from
dest_collection: The name of collection of where the data will be aggregated to
aggregation_name: The name for the published scheduled aggregation
description: The description for the published scheduled aggregation
aggregation_query: The aggregation query to schedule
date_field: The date field to check whether there is new data coming in
refresh_time: How often should the aggregation check for new data
start_immediately: Whether to start the published aggregation immediately

"""
		return requests.post(
			url=self.url+'/collection/publish_aggregation',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				source_collection=source_collection, 
				dest_collection=dest_collection, 
				aggregation_name=aggregation_name, 
				description=description, 
				aggregation_query=aggregation_query, 
				date_field=date_field, 
				refresh_time=refresh_time, 
				start_immediately=start_immediately, 
				))

	@retry()
	@return_curl_or_response('json')
	def delete_published_aggregation(self,aggregation_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/delete_published_aggregation',
			params=dict(
				aggregation_name=aggregation_name, 
				username=self.username, 
				api_key=self.api_key, 
				))

	@retry()
	@return_curl_or_response('json')
	def start_aggregation(self,aggregation_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/start_aggregation',
			params=dict(
				aggregation_name=aggregation_name, 
				username=self.username, 
				api_key=self.api_key, 
				))

	@retry()
	@return_curl_or_response('json')
	def stop_aggregation(self,aggregation_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/stop_aggregation',
			params=dict(
				aggregation_name=aggregation_name, 
				username=self.username, 
				api_key=self.api_key, 
				))

	@retry()
	@return_curl_or_response('json')
	def vector_aggregation(self, source_collection, dest_collection, source_to_dest_fields_mapping, vector_fields, aggregation_type="mean", refresh=True, **kwargs):
		"""Aggregate vectors from one collection into another published aggregation collection
This is useful for getting vectors of a category. e.g. You have "product\_description\_vector\_" and you want the vector for a brand samsung. The "samsung" brand's vector can be the aggregate of all the samsung "product\_description\_vector\_".
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
source_collection: The collection where the data to aggregate comes from
dest_collection: The collection that a scheduled aggregation is creating
source_to_dest_fields_mapping: The collection that a scheduled aggregation is creating
vector_fields: Vector fields to aggregate to form 1 aggregated vector for each split the groupby creates
aggregation_type: Aggregation for the vectors, choose from ['mean', 'sum', 'min', 'max']
refresh: Whether to refresh the aggregation and recalculate the vectors for every single groupby

"""
		return requests.post(
			url=self.url+'/collection/vector_aggregation',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				source_collection=source_collection, 
				dest_collection=dest_collection, 
				source_to_dest_fields_mapping=source_to_dest_fields_mapping, 
				vector_fields=vector_fields, 
				aggregation_type=aggregation_type, 
				refresh=refresh, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_array_field(self,array_fields, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/encode_array_field',
			params=dict(
				array_fields=array_fields, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_array(self,array_field, array, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/encode_array',
			params=dict(
				array_field=array_field, 
				array=array, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_with_array(self,array_field, array, collection_name, search_fields, approx=0, sum_fields=True, page_size=20, page=1, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		return requests.get(
			url=self.url+'/collection/search_with_array',
			params=dict(
				array_field=array_field, 
				array=array, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				search_fields=search_fields, 
				approx=approx, 
				sum_fields=sum_fields, 
				page_size=page_size, 
				page=page, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_dictionary_field(self,dictionary_fields, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/encode_dictionary_field',
			params=dict(
				dictionary_fields=dictionary_fields, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_dictionary(self, collection_name, dictionary, dictionary_field, **kwargs):
		"""Encode an dictionary into a vector
For example: a dictionary that represents a **person's characteristics visiting a store, field "person_characteristics"**:

    {"height":180, "age":40, "weight":70}

    -> <Encode the dictionary to vector> ->

| height | age | weight | purchases | visits |
|--------|-----|--------|-----------|--------|
| 180    | 40  | 70     | 0         | 0      |

    dictionary vector: [180, 40, 70, 0, 0]
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
dictionary: A dictionary to encode into vectors
dictionary_field: The dictionary field that encoding of the dictionary is trained on

"""
		return requests.post(
			url=self.url+'/collection/encode_dictionary',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				dictionary=dictionary, 
				dictionary_field=dictionary_field, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_with_dictionary(self, collection_name, search_fields, dictionary, dictionary_field, page_size=20, page=1, approx=0, sum_fields=True, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		"""Search a dictionary field with a dictionary using Vector Search
Vector similarity search with a dictionary directly.

For example: a dictionary that represents a **person's characteristics visiting a store, field "person_characteristics"**:

    {"height":180, "age":40, "weight":70}

    -> <Encode the dictionary to vector> ->

| height | age | weight | purchases | visits |
|--------|-----|--------|-----------|--------|
| 180    | 40  | 70     | 0         | 0      |

    dictionary vector: [180, 40, 70, 0, 0]

    -> <Vector Search> ->

    Search Results: {...}
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
search_fields: Vector fields to search against
page_size: Size of each page of results
page: Page of the results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
dictionary: A dictionary to encode into vectors
dictionary_field: The dictionary field that encoding of the dictionary is trained on

"""
		return requests.post(
			url=self.url+'/collection/search_with_dictionary',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				search_fields=search_fields, 
				page_size=page_size, 
				page=page, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				dictionary=dictionary, 
				dictionary_field=dictionary_field, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_fields_to_vector(self, collection_name, vector_name, selected_fields, **kwargs):
		"""Encode all selected fields for a collection into vectors
Within a collection encode the specified fields in every document into vectors.

For example: we choose the fields ["height", "age", "weight"]
    document 1 field: {"height":180, "age":40, "weight":70, "purchases":20, "visits": 12}

    document 2 field: {"height":160, "age":32, "weight":50, "purchases":10, "visits": 24}

    -> <Encode the fields to vectors> ->

| height | age | weight |
|--------|-----|--------|
| 180    | 40  | 70     | 
| 160    | 32  | 50     | 

    document 1 vector: {"person_characteristics_vector_": [180, 40, 70]}

    document 2 vector: {"person_characteristics_vector_": [160, 32, 50]}
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
vector_name: The name of the vector that the fields turn into
selected_fields: The fields to turn into vectors

"""
		return requests.post(
			url=self.url+'/collection/encode_fields_to_vector',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				vector_name=vector_name, 
				selected_fields=selected_fields, 
				))

	@retry()
	@return_curl_or_response('json')
	def encode_fields(self, collection_name, document, vector_name, **kwargs):
		"""Encode fields into a vector
For example: we choose the fields ["height", "age", "weight"]
    document field: {"height":180, "age":40, "weight":70, "purchases":20, "visits": 12}

    -> <Encode the fields to vectors> ->

| height | age | weight |
|--------|-----|--------|
| 180    | 40  | 70     |

    document vector: {"person_characteristics_vector_": [180, 40, 70]}

    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
document: A document to encode into vectors
vector_name: The name of the vector that the fields turn into

"""
		return requests.post(
			url=self.url+'/collection/encode_fields',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				document=document, 
				vector_name=vector_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def search_with_fields(self, collection_name, search_fields, document, selected_fields, vector_name, page_size=20, page=1, approx=0, sum_fields=True, metric="cosine", min_score=None, include_fields=[], include_vector=False, include_count=True, hundred_scale=False, asc=False, **kwargs):
		"""Search with fields with a document using Vector Search
Vector similarity search with fields directly.

For example: we choose the fields ["height", "age", "weight"]
    document field: {"height":180, "age":40, "weight":70, "purchases":20, "visits": 12}

    -> <Encode the fields to vectors> ->

| height | age | weight |
|--------|-----|--------|
| 180    | 40  | 70     |

    document dictionary vector: {"person_characteristics_vector_": [180, 40, 70]}

    -> <Vector Search> ->

    Search Results: {...}
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
search_fields: Vector fields to search against
page_size: Size of each page of results
page: Page of the results
approx: Used for approximate search
sum_fields: Whether to sum the multiple vectors similarity search score as 1 or seperate
metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
min_score: Minimum score for similarity metric
include_fields: Fields to include in the search results, empty array/list means all fields.
include_vector: Include vectors in the search results
include_count: Include count in the search results
hundred_scale: Whether to scale up the metric by 100
asc: Whether to sort results by ascending or descending order
document: A document to encode into vectors
selected_fields: The fields to turn into vectors
vector_name: A name to call the vector that the fields turn into

"""
		return requests.post(
			url=self.url+'/collection/search_with_fields',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				search_fields=search_fields, 
				page_size=page_size, 
				page=page, 
				approx=approx, 
				sum_fields=sum_fields, 
				metric=metric, 
				min_score=min_score, 
				include_fields=include_fields, 
				include_vector=include_vector, 
				include_count=include_count, 
				hundred_scale=hundred_scale, 
				asc=asc, 
				document=document, 
				selected_fields=selected_fields, 
				vector_name=vector_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def combine_vectors(self,vector_fields, vector_name, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/combine_vectors',
			params=dict(
				vector_fields=vector_fields, 
				vector_name=vector_name, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def collection_vector_mappings(self,collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/collection_vector_mappings',
			params=dict(
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def cluster(self,vector_field, collection_name, n_clusters=0, gpu=True, refresh=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/cluster',
			params=dict(
				vector_field=vector_field, 
				n_clusters=n_clusters, 
				gpu=gpu, 
				refresh=refresh, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def cluster_aggregate(self, collection_name, aggregation_query, page_size=20, page=1, asc=False, flatten=True, **kwargs):
		"""Aggregate every cluster in a collection
Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.

Only can be used after a vector field has been clustered with /cluster.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
aggregation_query: Aggregation query to aggregate data
page_size: Size of each page of results
page: Page of the results
asc: Whether to sort results by ascending or descending order
flatten: 

"""
		return requests.post(
			url=self.url+'/collection/cluster_aggregate',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				aggregation_query=aggregation_query, 
				page_size=page_size, 
				page=page, 
				asc=asc, 
				flatten=flatten, 
				))

	@retry()
	@return_curl_or_response('json')
	def cluster_facets(self,collection_name, facets_fields=[], page_size=1000, page=1, asc=False, date_interval="monthly", **kwargs):
		return requests.get(
			url=self.url+'/collection/cluster_facets',
			params=dict(
				facets_fields=facets_fields, 
				page_size=page_size, 
				page=page, 
				asc=asc, 
				date_interval=date_interval, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def cluster_centroids(self,vector_field, collection_name, **kwargs):
		return requests.get(
			url=self.url+'/collection/cluster_centroids',
			params=dict(
				vector_field=vector_field, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def cluster_centroid_documents(self,vector_field, collection_name, metric="cosine", include_vector=False, page=1, page_size=20, **kwargs):
		return requests.get(
			url=self.url+'/collection/cluster_centroid_documents',
			params=dict(
				vector_field=vector_field, 
				metric=metric, 
				include_vector=include_vector, 
				page=page, 
				page_size=page_size, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_cluster(self,vector_field, collection_name, alias="default", n_clusters=0, n_iter=10, n_init=5, gpu=True, refresh=True, **kwargs):
		return requests.get(
			url=self.url+'/collection/advanced_cluster',
			params=dict(
				vector_field=vector_field, 
				alias=alias, 
				n_clusters=n_clusters, 
				n_iter=n_iter, 
				n_init=n_init, 
				gpu=gpu, 
				refresh=refresh, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_cluster_aggregate(self, collection_name, aggregation_query, vector_field, alias, page_size=20, page=1, asc=False, flatten=True, filters=[], **kwargs):
		"""Aggregate every cluster in a collection
Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.

Only can be used after a vector field has been clustered with /advanced_cluster.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
aggregation_query: Aggregation query to aggregate data
page_size: Size of each page of results
page: Page of the results
asc: Whether to sort results by ascending or descending order
flatten: 
vector_field: Clustered vector field
alias: Alias of a cluster
filters: Query for filtering the search results

"""
		return requests.post(
			url=self.url+'/collection/advanced_cluster_aggregate',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				aggregation_query=aggregation_query, 
				page_size=page_size, 
				page=page, 
				asc=asc, 
				flatten=flatten, 
				vector_field=vector_field, 
				alias=alias, 
				filters=filters, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_cluster_facets(self,vector_field, collection_name, alias="default", facets_fields=[], page_size=1000, page=1, asc=False, date_interval="monthly", **kwargs):
		return requests.get(
			url=self.url+'/collection/advanced_cluster_facets',
			params=dict(
				vector_field=vector_field, 
				alias=alias, 
				facets_fields=facets_fields, 
				page_size=page_size, 
				page=page, 
				asc=asc, 
				date_interval=date_interval, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_cluster_centroids(self,vector_field, collection_name, alias="default", **kwargs):
		return requests.get(
			url=self.url+'/collection/advanced_cluster_centroids',
			params=dict(
				vector_field=vector_field, 
				alias=alias, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def advanced_cluster_centroid_documents(self,vector_field, collection_name, alias="default", metric="cosine", include_vector=False, page=1, page_size=20, **kwargs):
		return requests.get(
			url=self.url+'/collection/advanced_cluster_centroid_documents',
			params=dict(
				vector_field=vector_field, 
				alias=alias, 
				metric=metric, 
				include_vector=include_vector, 
				page=page, 
				page_size=page_size, 
				username=self.username, 
				api_key=self.api_key, 
				collection_name=collection_name, 
				))

	@retry()
	@return_curl_or_response('json')
	def insert_cluster_centroids(self, collection_name, cluster_centers, vector_field, alias="default", job=False, job_metric="cosine", **kwargs):
		"""Insert cluster centroids
Insert your own cluster centroids for it to be used in approximate search settings and cluster aggregations.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
cluster_centers: Cluster centers with the key being the index number
vector_field: Clustered vector field
alias: Alias is used to name a cluster
job: Whether to run a job where each document is assigned a cluster from the cluster_center
job_metric: Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']

"""
		return requests.post(
			url=self.url+'/collection/insert_cluster_centroids',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				cluster_centers=cluster_centers, 
				vector_field=vector_field, 
				alias=alias, 
				job=job, 
				job_metric=job_metric, 
				))

	@retry()
	@return_curl_or_response('json')
	def dimensionality_reduce(self, collection_name, vectors, vector_field, alias="default", n_components=1, **kwargs):
		"""Reduces the dimension of a list of vectors
Reduce the dimensions of a list of vectors you input into a desired dimension. 

This can only reduce to dimensions less than or equal to the n_components that the dimensionality reduction model is trained on.
    
Args
========
username: Username
api_key: Api Key, you can request it from request_api_key
collection_name: Name of Collection
vectors: Vectors to perform dimensionality reduction on
vector_field: Vector field to perform dimensionality reduction on
alias: Alias of the dimensionality reduced vectors
n_components: The size/length to reduce the vector down to.

"""
		return requests.post(
			url=self.url+'/collection/dimensionality_reduce',
			json=dict(
				username=self.username,
				api_key=self.api_key,
				collection_name=collection_name, 
				vectors=vectors, 
				vector_field=vector_field, 
				alias=alias, 
				n_components=n_components, 
				))




"""
Plotting functions of Vi go here. 
"""
import plotly.express as px
import plotly.graph_objects as go
import copy
import random
from typing import Union, List, Dict, Any
from plotly.subplots import make_subplots
from tqdm.notebook import tqdm
from .score import ViScore
from .utils import ViAnalyticsUtils, MeanDict
from ..read import ViReadClient
from ..errors import APIError

class VizMixin(ViScore, ViAnalyticsUtils):
    """
    Visualisation submodule for the library.
    """
    def _scatter_plot_documents(
        self,
        filtered_collection: list,
        dim_reduction_field: str,
        color: str,
        point_label: str,
        title: str=None,
        mode: str='markers',
        marker_size:int = 5
    ):
        """
        Add additional plot to figure and filtered collection and dim reduction field.

        Args:
            filtered_collection:
                A collection or the name of a collection
            dim_reduction_field:
                Dimensionally reduced field.
            color:
                Color of the plot
            point_label:
                Label of the plot
            title:
                Name of the group of documents

        Returns:
            go.Scatter: Scatter plot of centroids.

        """
        return go.Scatter(
            x=[
                self.get_field(dim_reduction_field, x)[0]
                for x in filtered_collection
            ],
            y=[
                self.get_field(dim_reduction_field, x)[1]
                for x in filtered_collection
            ],
            mode=mode,
            marker=dict(size=marker_size, color=color),
            text=[x[point_label] for x in filtered_collection],
            name=title,
        )

    def _scatter_plot_centroids(
        self,
        cluster_centroid_documents: List[Dict],
        dim_reduction_field: str,
        point_label: str,
        line_color="DarkSlateGrey",
        line_width=2,
        size=5,
        color_dict=None,
        mode: str='markers',
        color='orange',
    ):
        """
        Scatterplot the centroids

        Args:
            cluster_centroid_documents:
                The centroids to highlight.
            point_label:
                The label of every point. This should be found in the document.
            dim_reduction_field:
                The dimensionally-reduced vectors.
            line_width:
                The width of the lines.
            line_color:
                THe line color of the centroids
            color_dict:
                What to color the centroid the cluster label/title
            size:
                The size of the markers for the centroid documents
        
        Returns:
            go.Scatter: Scatter plot of the documents
        
        """
        x = [
            self.get_field(dim_reduction_field, x)[0]
            for x in cluster_centroid_documents.values()
        ]
        y = [
            self.get_field(dim_reduction_field, x)[1]
            for x in cluster_centroid_documents.values()
        ]
        titles = [
            self.get_field(point_label, x)
            for x in cluster_centroid_documents.values()
        ]
        return go.Scatter(
            x=x,
            y=y,
            mode=mode,
            marker=dict(
                size=size,
                # color=[color_dict[str(x['_clusters_'][cluster_field]['title'])] for x in cluster_centroids.values()],
                color=color,
                line=dict(width=line_width, color=line_color),
            ),
            text=titles,
            name="centroids",
        )

    def plot_dimensionality_reduced_vectors(
        self,
        collection: Union[str, List[Dict]],
        point_label: str,
        dim_reduction_field: str,
        cluster_field: str=None,
        cluster_label: str=None,
        include_centroids: bool = False,
        color: str=None,
        alias: str = None,
        mode: str='markers',
    ):
        """
        Returns a 2D plot of vectors that have been dimensionally reduced.

        Args:
            collection:
                A collection or the name of a collection
            point_label:
                The label of every point. This should be found in the document.
            dim_reduction_field:
                The dimensionally-reduced vectors.
            cluster_field:
                The field by which it is clustered.
            cluster_label:
                The name of the clusters
            include_centroids:
                Whether to include the centroids of every cluster
        
        Returns:
            Plotly figure object
        
        See example from: https://colab.research.google.com/drive/10u7b3lkIVJ-lceCmr34ywscOGueIFd0I?usp=sharing
        
        Example:
            >>> collection_name = 'nlp-qa'
            >>> cluster_field = 'question_vector_'
            >>> cluster_label = 'category'
            >>> alias = '1st_cluster'
            >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
            >>> vi_client.plot_dimensionality_reduced_vectors(collection=collection_name,
                    cluster_field=cluster_field,
                    cluster_label=cluster_label,
                    point_label='question_title',
                    dim_reduction_field=dim_reduction_field, 
                    include_centroids=True, 
                    alias=alias)
        
        """
        fig = go.Figure()
        if include_centroids:
            if not isinstance(collection, str):
                raise NotImplementedError(
                    "Centroid visualisation not included for custom clustering."
                    + "Please add them to yourself by using add_trace to the returned fig output."
                )

            cluster_centroid_documents = self.advanced_cluster_centroid_documents(
                collection, cluster_field, alias=alias
            )

            fig.add_trace(
                self._scatter_plot_centroids(
                    cluster_centroid_documents=cluster_centroid_documents,
                    dim_reduction_field=dim_reduction_field,
                    point_label=point_label,
                    mode=mode
                )
            )

        # If collection is simply a name, retrieve collection
        if isinstance(collection, str):
            fields_to_include = [point_label, dim_reduction_field, cluster_field, cluster_label]
            fields_to_include = [x for x in fields_to_include if x is not None]
            collection = self.retrieve_all_documents(collection, include_fields=[fields_to_include])

        if cluster_label is None:
            color_dict = None
        else:
            cluster_titles = [
                self.get_field(cluster_label, x) for x in collection
            ]
            cluster_titles = list(set(cluster_titles))
            cluster_titles_color_map = dict(enumerate(cluster_titles))
            color_dict = {v: k + 5 for k, v in cluster_titles_color_map.items()}
        
        if cluster_label is None:
            fig.add_trace(
                self._scatter_plot_documents(
                    filtered_collection=collection,
                    dim_reduction_field=dim_reduction_field,
                    color=color,
                    point_label=point_label,
                    mode=mode
                )
            )
        else:
            for title in tqdm(cluster_titles):
                filtered_collection = [
                    x
                    for x in collection
                    if self.get_field(cluster_label, x) == title
                ]
                fig.add_trace(
                    self._scatter_plot_documents(
                        title=title,
                        filtered_collection=filtered_collection,
                        dim_reduction_field=dim_reduction_field,
                        color=color_dict[title],
                        point_label=point_label,
                    )
                )
            fig.update_layout(showlegend=True)
        return fig

    def plot_1d_cosine_similarity(
        self,
        documents: list,
        vector_fields: List[str],
        label: str,
        anchor_document: dict = None,
        anchor_index: Union[str, int] = 0,
        orientation: int="h",
        barmode: int="group",
        num_cols: int=None,
        y_axis_tickangle: int=-15,
        x_axis_tickangle: int=15
    ):
        """
        Compare 1 document against other documents.
        Ensure that the name is unique, otherwise the plot will simply take the mean.
        
        Args:
            documents: list of documents (dictionaries) to feed in
            vector_fields: vector field to calculate cosine similarity on 
            label: the x label for the bar plot 
            anchor_document: the document to compare it on
            anchor_index: the anchor index to compare it on
            orientation: The orientation of the bar chart. Can be 'v' (vertical) or 'h' (horizontal)
            num_cols: The number of columns. The default will put everything into 1 row. If you want to 
            put things into multiple rows, then please reduce the number of columns. 
            y_axis_tickangle: This will change the tick angles of the y axis in the vertical chart.
            x_axis_tickangle: This will change the tick angels of the x axis in the vertical chart.
        Returns:
            Plotly Figure: returns a horizontal barplot showing cosine similarity scores.
        
        Example:
            >>> cluster_field = 'question_vector_'
            >>> cluster_label = 'category'
            >>> alias = '1st_cluster'
            >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
            >>> vi_client.plot_1d_cosine_similarity(collection=collection_name,
                    cluster_field=cluster_field,
                    cluster_label=cluster_label,
                    point_label='question_title',
                    dim_reduction_field=dim_reduction_field, 
                    include_centroids=True, 
                    alias=alias)
        
        """
        if isinstance(vector_fields, str):
            vector_fields = [vector_fields]
        if anchor_document is None:
            other_documents = copy.deepcopy(documents)
            if isinstance(anchor_index, int):
                anchor_document = other_documents.pop(anchor_index)
            elif isinstance(anchor_index, str):
                # Loop through list to get ID
                for i, doc in enumerate(other_documents):
                    if anchor_index == doc["_id"]:
                        anchor_document = other_documents.pop(i)
                        break
        import math
        if num_cols is None:
            num_cols = len(vector_fields)
        fig = make_subplots(rows=int(len(vector_fields) / num_cols) + 1, cols=num_cols, 
        subplot_titles=vector_fields)
        row_num = 1
        col_num = 1
        for i, vector_field in enumerate(vector_fields):
            fig.add_trace(self._plot_1d_cosine_similarity_for_1_vector_field(
                documents=documents,
                vector_field=vector_field,
                label=label,
                anchor_document=anchor_document, 
                anchor_index=anchor_index,
                orientation=orientation),
                row=row_num, col=col_num)
            if col_num % num_cols == 0:
                row_num += 1
                col_num = 1
            else:
                col_num += 1
        if orientation == 'h':
            fig.update_yaxes(tickangle=y_axis_tickangle)
        if orientation == 'v':
            fig.update_xaxes(tickangle=x_axis_tickangle)
        
        return fig
    
    def _plot_1d_cosine_similarity_for_1_vector_field(self, 
        documents: list,
        vector_field: str,
        label: str,
        anchor_document: dict=None,
        anchor_index: Union[str, int]=0,
        orientation="h") -> go.Bar:
            """
            Compare 1 document against other documents.

            Args:
                documents: list of documents (dictionaries) to feed in
                vector_fields: vector field to calculate cosine similarity on 
                label: the x label for the bar plot 
                anchor_document: the document to compare it on
                anchor_index: the anchor index to compare it on
                orientation: The orientation of the bar chart
            Returns:
                None, returns a horizontal barplot showing cosine similarity scores.

            Example:
                >>> cluster_field = 'question_vector_'
                >>> cluster_label = 'category'
                >>> alias = '1st_cluster'
                >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
                >>> vi_client.plot_1d_cosine_similarity(collection=collection_name,
                        cluster_field=cluster_field,
                        cluster_label=cluster_label,
                        point_label='question_title',
                        dim_reduction_field=dim_reduction_field, 
                        include_centroids=True, 
                        alias=alias)
            
            """
            
            try:
                scores = self.get_cosine_similarity_scores(
                    documents, anchor_document, vector_field=vector_field
                )
            except NameError:
                raise APIError(f"Anchor document with {anchor_index} not found.")

            for i, doc in enumerate(documents):
                doc["cos_score"] = scores[i]

            mean_dict = MeanDict()
            for i, doc in enumerate(documents):
                mean_dict[doc[label]] = doc["cos_score"]
            
            x, y = mean_dict.get_x_y()
            if orientation == 'h':
                return go.Bar(
                    x=y,
                    y=x,
                    orientation=orientation,
                    name=vector_field
                )
            else:
                return go.Bar(
                    x=x,
                    y=y,
                    text=x,
                    orientation=orientation,
                    name=vector_field
                )
        
    # def plot_2d_cosine_similarity(
    #     self,
    #     documents: List[Dict[str, Any]],
    #     anchor_documents: List[Dict[str, Any]],
    #     vector_fields: Union[str, List[str]],
    #     label: str,
    #     mode: str='markers+text'
    # ):
    #     """
    #     Plot cosine similarity with 2 anchor documents to compare with against other documents.

    #     Args:
    #         documents: list of documents (dictionaries) to feed in
    #         vector_field: vector field to calculate cosine similarity on 
    #         label: the x label for the bar plot 
    #         anchor_document: the document to compare it on
    #         anchor_index: the anchor index to compare it on
    #     Returns:
    #         A scatterplot showing cosine similarity scores with each axes being a specific document.
        
    #     Example:
    #         >>> cluster_field = 'question_vector_'
    #         >>> cluster_label = 'category'
    #         >>> alias = '1st_cluster'
    #         >>> dim_reduction_field = '_dr_.default.2.question_vectors_' # the '.' in names implies nested dictionaries in Vi
    #         >>> vi_client.plot_2d_cosine_similarity(documents=documents[2:],
    #                 anchor_documents=documents[:2],
    #                 vector_fields=vector_field,
    #                 label=label
    #                 )

    #     """
    #     assert (
    #         len(anchor_documents) == 2
    #     ), "You need 2 anchor documents for a 2d cosine similarity plot."
    #     fig = go.Figure()
    #     if isinstance(vector_fields, str):
    #         vector_fields = [vector_fields]
    #     for vector_field in vector_fields:
            
    #         scores_x = self.get_cosine_similarity_scores(
    #             documents, anchor_documents[0], vector_field
    #         )
    #         scores_y = self.get_cosine_similarity_scores(
    #             documents, anchor_documents[1], vector_field
    #         )

    #         for i, doc in enumerate(documents):
    #             doc["cos_score_x"] = scores_x[i]
    #             doc["cos_score_y"] = scores_y[i]

    #         x = [round(x["cos_score_x"], 3) for x in documents]
    #         y = [round(x["cos_score_y"], 3) for x in documents]
    #         labels = [x[label] for x in documents]
    #         fig.add_trace(go.Scatter(x=x, y=y, text=labels, mode=mode, name=vector_field))
    #     fig.update_xaxes(title_text=f"Comparing with {anchor_documents[0][label]}")
    #     fig.update_yaxes(title_text=f"Comparing with {anchor_documents[1][label]}")
    #     fig.update_layout(
    #         title_text=f"2D Cosine Similarity Comparison With {anchor_documents[0][label]} and {anchor_documents[1][label]}"
    #     )
    #     return fig

    def random_colour(self, num_of_colors) -> List:
        levels = range(32,256,32)
        return [tuple(random.choice(levels) for _ in range(3)) for _ in range(num_of_colors)]

    def add_line_to_fig(self, fig, x0: float=0, y0: float=0, x1: float=1, y1: float=1, color: str='LightSeaGreen', width: float=1, dash: str="dashdot"):
        fig.add_shape(type="line",
            x0=x0, y0=y0, x1=x1, y1=y1,
            line=dict(
                color=color,
                width=width,
                dash=dash,
            )
        )
        return fig

    def plot_2d_cosine_similarity(self, 
        documents: List[Dict], anchor_documents: List[Dict], vector_fields: List[str], 
        label: str, mode='markers+text', textposition='top center', show_spikes=True,
        text_label_font_size: int=12, text_label_font_family = "Rockwell",
        text_label_bgcolor="white",
        marker_colors=['purple', 'aquamarine'], metric='cosine',
        plot_bgcolor="#e6e6fa", spikedash='dot', spikethickness=1.5, include_diagonal_line: bool=True
    ):
        """
        Plotting 2D cosine similarity plots 
        Args:
            documents: The documents to 
            anchor_documents: Documents by which to compare against
            vector_fields: The list of vectors to accept 
            Label: The document field to label
            Mode: Whether to include markers or text (view Plotly documentation for more information)
            textposition: where the text labels should be in relation to the marker
            show_spikes: show the spikes in comparison to the x and y labels
            text_label_font_family: The font of the text
            text_label_font_size: The font size of the text
            marker_colors: The color of the markers If the number of colors do not
            match then we it randomly generates.
            metric: The metric to use. Currently only supports cosine similarity
            plot_bgcolor: The background color of the plot
            spikethickness: The thickness of the spikes
            spikedash: Type of line the spikes should be.
            include_diagonal_line: Include the diagonal line.
        Example:
            >>> vi_client = ViClient()
            >>> collection_name = 'ecommerce'
            >>> docs = vi_client.random_documents(collection_name)['documents']
            >>> vi_client.plot_2d_cosine_similarity(docs, docs[0:2], 
            vector_fields=['use_vector_'], label='name')
        """
        if metric != 'cosine':
            raise NotImplementedError("Cosine similarity score is currently not implemented.")

        assert (
            len(anchor_documents) == 2
        ), "You need 2 anchor documents for a 2d cosine similarity plot."

        fig = go.FigureWidget()
        
        if len(vector_fields) > len(marker_colors):
            num_of_extra_fields = len(vector_fields) - len(marker_colors)
            marker_colors += self.random_colour(num_of_extra_fields)

        for vector_field_counter, vector_field in enumerate(vector_fields):
            scores_x = self.get_cosine_similarity_scores(
                documents, anchor_documents[0], vector_field
            )
            scores_y = self.get_cosine_similarity_scores(
                documents, anchor_documents[1], vector_field
            )

            for i, doc in enumerate(documents):
                doc["cos_score_x"] = scores_x[i]
                doc["cos_score_y"] = scores_y[i]
            x = [round(x["cos_score_x"], 3) for x in documents]
            y = [round(x["cos_score_y"], 3) for x in documents]
            labels = self.get_field_across_documents(label, documents)
            comparisons = [x_i > y[i] for i, x_i in enumerate(x)]
            text_comparisons = ["has <b>lower</b> cosine similarity with" if c == True 
            else "has <b>higher</b> cosine similarity with" for c in comparisons]
            fig.add_trace(go.Scatter(x=x, y=y, 
                text=labels,
                mode=mode, 
                name=vector_field,
                customdata=text_comparisons,
                hovertemplate = label +": <b>%{text}<extra></extra><br></b>" + \
                "%{customdata} <br>" + \
                "<b>" + anchor_documents[1][label] + "</b> (%{y})<br>" + \
                "compared to <br>" + \
                "<b>" + anchor_documents[0][label] + "</b> (%{x})</b>",
                marker=dict(
                    color=marker_colors[vector_field_counter],
                    size=5,
                    line=dict(width=0.5, color='DarkSlateGrey')
                )))

        self.add_labels_to_figure(fig, x_axis_label=f"Cosine Similarity With {anchor_documents[0][label]}",
        y_axis_label=f"Cosine Similarity WIth {anchor_documents[1][label]}",
        title_text=f"2D Cosine Similarity Comparison With {anchor_documents[0][label]} and {anchor_documents[1][label]}")

        fig.update_traces(textposition=textposition)

        fig.update_xaxes(showspikes=show_spikes, 
        spikedash=spikedash, spikethickness=spikethickness)
        fig.update_yaxes(showspikes=show_spikes, 
        spikedash=spikedash, spikethickness=spikethickness)
        
        fig.update_layout(plot_bgcolor=plot_bgcolor)
        fig.update_layout(
            hoverlabel=dict(
                bgcolor=text_label_bgcolor,
                font_size=text_label_font_size,
                font_family=text_label_font_family
            )
        )
        if include_diagonal_line:
            fig = self.add_line_to_fig(fig, x0=0, y0=0, x1=1, y1=1)
        return fig
    
    def add_labels_to_figure(self, fig, title_text: str, x_axis_label: str, y_axis_label: str):
        fig.update_xaxes(title_text=x_axis_label)
        fig.update_yaxes(title_text=y_axis_label)
        fig.update_layout(title_text=title_text)

    def _plot_radar(self, scores: list, spokes: list, name: str, fill=None):
        """
            Args:
                Scores: The list of scores for cosine similarity
                spokes: The outside labels 
                name: The name of the plot
        """
        return go.Scatterpolar(
            # Get the cosine similarity scores here 
            r=scores + [scores[0]],
            theta=spokes + [spokes[0]],
            fill=fill,
            name=name,
        )

    def plot_radar_across_documents(self, docs: List[Dict], anchor_documents: List[Dict], vector_field: str, 
    label_field: str, range: List=[0, 1], fill: str=None, scoring_metric ='cosine'):
        """
            Radar plot for 1D cosine similarity across documents.
            Args:
                docs: A list of documents 
                anchor_document: The document to compare against
                vector_field: The vector vector field
                label_field: The field of the documents to get labels.
        """
        categories = self.get_field_across_documents(label_field, docs)
        fig = go.Figure()
        for anchor_document in anchor_documents:
            if scoring_metric == 'cosine':
                scores = self.get_cosine_similarity_scores(docs, anchor_document, vector_field=vector_field)
            else:
                scores = scoring_metric(docs, anchor_document, vector_field=vector_field)
            fig.add_trace(self._plot_radar(scores=scores, spokes=categories, name=vector_field, fill=fill))
        
        fig.update_layout(
            polar=dict(
            radialaxis=dict(
                visible=True,
                range=range
            )),
            showlegend=True
        )
        return fig

    def plot_radar_across_vector_fields(self, docs: List[Dict], anchor_document: Dict, 
    vector_fields: List[str], label_field: str, range=[0, 1], fill=None, scoring_metric='cosine'):
        """
            Radar plot for 1D cosine similarity across different vector spaces.
            Args:
                docs: A list of documents 
                anchor_document: The document to compare against
                vector_fields: the different vector fields
                label_field: The field of the documents to get labels.
        """
        categories = self.get_field_across_documents(label_field, docs)
        fig = go.Figure()
        for vector in vector_fields:
            if scoring_metric == 'cosine':
                scores = self.get_cosine_similarity_scores(docs, anchor_document, vector_field=vector)
            else:
                scores = scoring_metric(docs, anchor_document, vector_field=vector)
            
            fig.add_trace(self._plot_radar(scores=scores, spokes=categories, name=vector, fill=fill))

        fig.update_layout(
            polar=dict(
            radialaxis=dict(
                visible=True,
                range=range
            )),
            showlegend=True 
        )
        return fig
